#ifndef overlay_window_h_included
#define overlay_window_h_included

#define NOMINMAX
#define WIN32_LEAN_AND_MEAN
#include <Windows.h>

#include "resource.h"

#include <vc_base.hpp>

#include <managed_window.hpp>
#include <bitmap.hpp>
#include <clip_region.hpp>

#include <algorithm> // for std::transform
#include <iterator> // for std::back_inserter
#include <vector>


#define WM_USER_VIEWPORT_CHANGED WM_USER + 1


class overlay_window : public managed_window {
private:

    bitmap _img; /// the background image to display
    HWND _hwndOriginalParent; /// the parent window for the overlay
    RECT _rcOriginalPosition; /// the last received position of the overlay
    std::vector<RECT> _clipRects;
    vcsdk::viewport* _viewport;

public:

    /// constructor. default constructor.
    overlay_window(HWND hndl, DeleterFn&& deleter, vcsdk::viewport& viewport)
        : managed_window(hndl, std::move(deleter))
        , _img(bitmap::from_resource(get_app_instance(), IDB_BACKGROUND))
        , _hwndOriginalParent(NULL)
        , _rcOriginalPosition({})
        , _viewport(&viewport) {
    }

    /// constructor. move constructor.
    overlay_window(overlay_window&&) = default;

    /// move assignment operator.
    overlay_window& operator =(overlay_window&&) = default;

    /// overlay window factory method
    template <typename Container>
    static bool emplace_new(Container& c, const typename Container::key_type& key, const std::wstring& title, HINSTANCE hInstance, vcsdk::viewport& viewport) {
        using window_class_type = window_class<overlay_window>;
        static window_class_type wndclass = window_class_type::register_class(hInstance, L"ctxoverlay_window_class2");

        return managed_window::emplace_new(c, key, wndclass, title, viewport);
    }

    void reset_image(bitmap&& img) { _img = std::move(img); redraw(); }
    void reset_image() { reset_image(bitmap::from_resource(get_app_instance(), IDB_BACKGROUND)); }

    /// return the original cached parent for the window.
    HWND get_original_parent() { return _hwndOriginalParent; }

    /// reparent the overlay window.
    void reparent(HWND hwndParent) {

        embed_into((HWND)_viewport->translate_app_window(hwndParent));

        // save original parent
        _hwndOriginalParent = hwndParent;
    }

    /// return the original cached position of the window in VDI/server coordinates.
    RECT get_original_position() { return _rcOriginalPosition; }

    /// reposition the overlay window. the rectangle provided is in VDI/server coordinates.
    void reposition(RECT rc) {

        if (_viewport->is_seamless()) {
            MapWindowPoints(HWND_DESKTOP, get_parent(), (LPPOINT)&rc, 2);
        }

        move_window(_viewport->translate_virtual_rect(rc, VDTF_WINDOW_RECT));

        // save original position
        _rcOriginalPosition = rc;
    }

    /// return the original cached clipping rectangles in VDI/server coordinates.
    std::vector<RECT>& get_original_clip_rects() { return _clipRects; }

    /// apply clipping rectangles to the overlay window.
    void clip(std::vector<RECT>& rects) {

        std::vector<RECT> translated_rects;
        std::transform(rects.begin(), rects.end(), std::back_inserter(translated_rects), [&](RECT rc) {
            return _viewport->translate_virtual_rect(rc, VDTF_CLIENT_AREA_RECT);
            });

        // save original clipping rects
        _clipRects = rects;

        if (translated_rects.size()) {
            clip_to_region(clip_region::from_rectangle_list(translated_rects));
        } else {
            clear_region();
        }
    }


    void on_viewport_changed() {

        // recalculate parent from original parent
        reparent(get_original_parent());

        // recalculate position from original position
        reposition(get_original_position());

        // recalculate clipping rectangles from original clipping rectangles
        clip(get_original_clip_rects());
    }


    LRESULT on_message(UINT message, WPARAM wparam, LPARAM lparam, bool& handled) {
        switch (message) {
        case WM_PAINT:
            handled = on_paint();
            break;

        case WM_USER_VIEWPORT_CHANGED:
            on_viewport_changed();
            break;

        default:
            relay_message(message, wparam, lparam);
            break;
        }

        return 0;
    }

private:

    /// relays window messages to the virtual desktop application.
    void relay_message(UINT message, WPARAM wparam, LPARAM lparam) {

        // pass keyboard and mouse messages to parent window. using the WS_EX_TRANSPARENT style
        // should do this automatically but seems to cause flicker and didnt work.

        if ((message >= WM_MOUSEFIRST && message <= WM_MOUSELAST) ||
            (message >= WM_KEYFIRST && message <= WM_KEYLAST) ||
            (message == WM_MOUSEACTIVATE)) {

            if (HWND hwndParent = GetParent(_hwnd); hwndParent && IsWindow(hwndParent)) {

                if (message >= WM_MOUSEFIRST && message <= WM_MOUSELAST) {

                    // coordinates for mouse messages are client coordinates. we need to adjust them
                    // to account for the offset from the parent's top left corner.

                    POINT point = {};
                    point.x = LOWORD(lparam);
                    point.y = HIWORD(lparam);

                    MapWindowPoints(_hwnd, hwndParent, &point, 1);

                    lparam = MAKELPARAM(point.x, point.y);
                }

                PostMessage(hwndParent, message, wparam, lparam);
            }
        }
    }

    /// draw the background image in the window
    bool on_paint() {
        paint_all([this](device_context& dc, RECT rc) {
            if (_img.valid()) { dc.draw(_img, rc); }
            draw_border(dc, rc, RGB(255, 0, 255));
            });

        return true;
    }

    void draw_border(device_context& dc, const RECT& rc, COLORREF color) {
        HBRUSH red_brush = CreateSolidBrush(color);
        dc.using_object(red_brush, [&red_brush, &rc](device_context& dc) {
            RECT rcBorder{};
            CopyRect(&rcBorder, &rc);
            InflateRect(&rcBorder, -1, -1);
            dc.frame_rect(rcBorder, red_brush);
            });

        DeleteObject(red_brush);
    }
};


#endif