/*************************************************************************
*
* Copyright (C) Citrix Systems, Inc. All Rights Reserved.
*
* vdoverlay.cpp
*
* Virtual Driver Example - Overlays
*
* $Id: 
*
* The code and comments contained in this file are provided "as is"
* without warranty of any kind, either expressed or implied,
* including, but not limited to, the implied warranties of
* merchantability and/or fitness for a particular purpose.
* You have the right to use, modify, reproduce, and distribute
* this file, and/or any modified version of this file, in any way
* that you find useful, provided that you agree that Citrix Systems
* has no warranty obligations or liability for this or any other
* sample file.
*
*************************************************************************/

#include "citrix.h" // includes windows.h automatically

#undef min // defined in platcomm.h which is included from citrix.h, conflicts with std::min
#undef max // defined in platcomm.h which is included from citrix.h, conflicts with std::max

#include "overlay_vc_controller.hpp"
#include "overlay_vc.hpp"
#include "appshare_vc.hpp"

#include <vdoverlay.h>

/* from src/inc */
#include "logapi.h"
#include "miapi.h"





HINSTANCE hDll = NULL;

BOOL APIENTRY DllMain(HANDLE hModule, DWORD ul_reason_for_call, LPVOID ignore)
{
    switch (ul_reason_for_call) {
    case DLL_PROCESS_ATTACH:
        hDll = (HINSTANCE)hModule;
    }
    return TRUE;
}


/*******************************************************************************
 *
 *  DriverOpen
 *
 *    Called once to set up things.
 *
 * ENTRY:
 *    pVd (input)
 *       pointer to virtual driver data structure
 *    pVdOpen (input/output)
 *       pointer to the structure VDOPEN
 *    puiSize (Output)
 *       size of VDOPEN structure.
 *
 * EXIT:
 *    CLIENT_STATUS_SUCCESS - no error
 *    CLIENT_ERROR_NO_MEMORY - could not allocate data buffer
 *    On other errors, an error code is returned from lower level functions.
 *
 ******************************************************************************/

int DriverOpen(PVD pVd, PVDOPEN pVdOpen, PUINT16 puiSize)
{
    try {
        // there is no need to have app-sharing and overlay management implemented in separate virtual channels.
        // this was a deliberate design decision to demonstrate that it can be done, not necessarily a recommendation
        // or endorsement of this particular design choice.
        overlay_vc_controller* controller = new overlay_vc_controller(pVd, CTXOVERLAY_VER_LO, CTXOVERLAY_VER_HI, hDll);
        controller->open_vc<overlay_virtual_channel>(CTXOVERLAY_VIRTUAL_CHANNEL_NAME);

        if (VdIsAppProtectionEnabled(pVd) == FALSE) {
            controller->open_vc<appshare_virtual_channel>(CTXAPPSHARE_VIRTUAL_CHANNEL_NAME);
        }
    }
    catch (const vc_exception& ex) {
        return ex.code().value();
    }

    return(CLIENT_STATUS_SUCCESS);
}


/*******************************************************************************
 *
 *  DriverPoll
 *
 *  The Winstation driver calls DriverPoll
 *
 * ENTRY:
 *    pVd (input)
 *       pointer to virtual driver data structure
 *    pVdPoll (input)
 *       pointer to the structure DLLPOLL or DLL_HPC_POLL
 *    puiSize (input)
 *       size of DLLPOOL structure.
 *
 * EXIT:
 *    CLIENT_STATUS_SUCCESS - OK so far.  We will be polled again.
 *    CLIENT_STATUS_NO_DATA - No more data to send.
 *    CLIENT_STATUS_ERROR_RETRY - Could not send the data to the WD's output queue - no space.
 *                                Hopefully, space will be available next time.
 *    Otherwise, a fatal error code is returned from lower level functions.
 *
 *    NOTE:  CLIENT_STATUS_NO_DATA signals the WD that it is OK to stop
 *           polling until another host to client packet comes through.
 *
 * REMARKS:
 *    If polling is enabled (pre-HPC client, or HPC client with polling enabled),
 *    this function is called regularly.  DriverPoll is always called at least once.
 *    Otherwise (HPC with polling disabled), DriverPoll is called only when requested
 *    via WDSET_REQUESTPOLL or SENDDATA_NOTIFY.
 *
 ******************************************************************************/

int DriverPoll(PVD pVd, PDLLPOLL pVdPoll, PUINT16 puiSize)
{
    try {
        return vcsdk::vc_controller::from_handle(pVd)->poll() ? CLIENT_STATUS_SUCCESS : CLIENT_STATUS_NO_DATA;
    }
    catch (const vc_exception& ex) {
        return ex.code().value();
    }
}

/*******************************************************************************
 *
 *  DriverClose
 *
 *  The user interface calls VdClose to close a Vd before unloading it.
 *
 * ENTRY:
 *    pVd (input)
 *       pointer to procotol driver data structure
 *    pVdClose (input/output)
 *       pointer to the structure DLLCLOSE
 *    puiSize (input)
 *       size of DLLCLOSE structure.
 *
 * EXIT:
 *    CLIENT_STATUS_SUCCESS - no error
 *
 ******************************************************************************/

int DriverClose(PVD pVd, PDLLCLOSE pVdClose, PUINT16 puiSize)
{
    try {
        delete vcsdk::vc_controller::from_handle(pVd);
    }
    catch (const vc_exception& ex) {
        return ex.code().value();
    }

    return(CLIENT_STATUS_SUCCESS);
}

/*******************************************************************************
 *
 *  DriverInfo
 *
 *    This routine is called to get module information
 *
 * ENTRY:
 *    pVd (input)
 *       pointer to virtual driver data structure
 *    pVdInfo (output)
 *       pointer to the structure DLLINFO
 *    puiSize (output)
 *       size of DLLINFO structure
 *
 * EXIT:
 *    CLIENT_STATUS_SUCCESS - no error
 *
 ******************************************************************************/

int DriverInfo(PVD pVd, PDLLINFO pVdInfo, PUINT16 puiSize)
{
    *puiSize = sizeof(DLLINFO);

    try {
        return vcsdk::vc_controller::from_handle(pVd)->info(pVdInfo->pBuffer, &pVdInfo->ByteCount);
    }
    catch (const vc_exception& ex) {
        return ex.code().value();
    }

    return CLIENT_STATUS_SUCCESS;
}

/*******************************************************************************
 *
 *  DriverQueryInformation
 *
 *   Required vd function.
 *
 * ENTRY:
 *    pVd (input)
 *       pointer to virtual driver data structure
 *    pVdQueryInformation (input/output)
 *       pointer to the structure VDQUERYINFORMATION
 *    puiSize (output)
 *       size of VDQUERYINFORMATION structure
 *
 * EXIT:
 *    CLIENT_STATUS_SUCCESS - no error
 *
 ******************************************************************************/

int DriverQueryInformation(PVD pVd, PVDQUERYINFORMATION pVdQueryInformation, PUINT16 puiSize)
{
    *puiSize = sizeof(VDQUERYINFORMATION);

    return(CLIENT_STATUS_SUCCESS);
}

/*******************************************************************************
 *
 *  DriverSetInformation
 *
 *   Required vd function.
 *
 * ENTRY:
 *    pVd (input)
 *       pointer to virtual driver data structure
 *    pVdSetInformation (input/output)
 *       pointer to the structure VDSETINFORMATION
 *    puiSize (input)
 *       size of VDSETINFORMATION structure
 *
 * EXIT:
 *    CLIENT_STATUS_SUCCESS - no error
 *
 ******************************************************************************/

int DriverSetInformation(PVD pVd, PVDSETINFORMATION pVdSetInformation, PUINT16 puiSize)
{
    try {
        vcsdk::vc_controller::from_handle(pVd)->set_info(pVdSetInformation, puiSize);
    }
    catch (const vc_exception& ex) {
        return ex.code().value();
    }

    return(CLIENT_STATUS_SUCCESS);
}

/*******************************************************************************
 *
 *  DriverGetLastError
 *
 *   Queries error data.
 *   Required vd function.
 *
 * ENTRY:
 *    pVd (input)
 *       pointer to virtual driver data structure
 *    pLastError (output)
 *       pointer to the error structure to return (message is currently always
 *       NULL)
 *
 * EXIT:
 *    CLIENT_STATUS_SUCCESS - no error
 *
 ******************************************************************************/

int DriverGetLastError(PVD pVd, PVDLASTERROR pLastError)
{
    // Interpret last error and pass back code and string data

    pLastError->Error = pVd->LastError;
    pLastError->Message[0] = '\0';
    return(CLIENT_STATUS_SUCCESS);
}



/*******************************************************************************
 *
 *  ICADataArrival
 *
 *   A data PDU arrived over our channel.
 *
 * ENTRY:
 *    pVd (input)
 *       pointer to virtual driver data structure
 *
 *    uChan (input)
 *       ICA channel the data is for.
 *
 *    pBuf (input)
 *       Buffer with arriving data packet
 *
 *    Length (input)
 *       Length of the data packet
 *
 * EXIT:
 *       void
 *
 ******************************************************************************/

void WFCAPI ICADataArrival(PVD pVd, USHORT uChan, LPBYTE pBuf, USHORT Length)
{
    try {
        MEMORY_SECTION buf{};
        buf.pSection = pBuf;
        buf.length = Length;
        vcsdk::vc_controller::from_handle(pVd)->receive_data(uChan, buf);
    }
    catch (const vc_exception& ex) { //set last error?
        UNREFERENCED_PARAMETER(ex);
    }
}
