#pragma once

#include "overlay_content.hpp"

#include <appshare.h>

#include <memory> // for std::unique_ptr
#include <functional> // for std::function


class app_sharing_session {
private:

	static session_event convert_event(WFAppDisplayEvent evt) {
		switch (evt) {
		case WFAppDisplayStarted:
			return session_event::started;

		case WFAppDisplayUpdated:
			return session_event::updated;

		case WFAppDisplayStopped:
		default:
			return session_event::stopped;
		}
	}

	// begin{
	// restructure the WFAPI SDK app-sharing APIs to work with the unique_resource idiom. the deleter function should accept what the acquisition function returns as a parameter.
	static void WINAPI capture_event(WFAppDisplayEvent evt, void* data, void* client) {
		if (callback_type* callback = reinterpret_cast<callback_type*>(client); callback && *callback) {
			(*callback)(convert_event(evt), reinterpret_cast<WFAppDisplayInfo*>(data)->bufferid);
		}
	}

	static window begin_capture(const window& win, void* client, UINT32 flags) {
		return (WFCaptureAppDisplay(win, &capture_event, client, flags) == TRUE) ? win : window{};
	}

	static void end_capture(const window& win) {
		WFReleaseAppDisplay(win);
	}
	// restructure apis...
	// end}

	struct release {
		using pointer = window;
		void operator ()(pointer hndl) const { end_capture(hndl); }
	};

	using callback_type = std::function<void(session_event evt, const unsigned int& data)>;
	using unique_resource_type = std::unique_ptr<window, release>; /// managed resource type

	std::unique_ptr<callback_type> _callback; /// heap allocated function callback for processing app-sharing session events.
	unique_resource_type _session; /// unique resource to manage the lifecycle of the app-sharing session.

public:

	/// the session key type.
	using key_type = HWND;

	static app_sharing_session create(const window& app_window, callback_type&& callback, UINT32 flags) {
		return app_sharing_session(app_window, std::make_unique<callback_type>(std::move(callback)), flags);
	}

	/// constructor. default constructor.
	app_sharing_session() = default;

	/// constructor.
	app_sharing_session(std::unique_ptr<callback_type>&& callback, unique_resource_type&& session)
		: _callback(std::move(callback))
		, _session(std::move(session)) {
	}

	/// constructor. construct with application window, callback and capture flags.
	app_sharing_session(const window& app_window, std::unique_ptr<callback_type>&& callback, UINT32 flags)
		: app_sharing_session(std::move(callback), unique_resource_type(begin_capture(app_window, callback.get(), flags))) {
	}

	/// copy constructor. copy construction is not allowed.
	app_sharing_session(const app_sharing_session&) = delete;

	/// copy assignment operator. copy assignment is not allowed.
	app_sharing_session& operator =(const app_sharing_session&) = delete;

	/// move constructor.
	app_sharing_session(app_sharing_session&&) = default;

	/// move assignment operator.
	app_sharing_session& operator =(app_sharing_session&&) = default;

	/// destructor.
	~app_sharing_session() {
		_session.reset(); // manual reset here to ensure that the session is deleted before the callback since the session has a pointer to the callback.
	}
};

/// live app-sharing content implementation. managed resource: the underlying app display session created by WFCaptureAppDisplay
using application_content = overlayed_content<app_sharing_session, WM_USER_SUBSCRIBEAPP>;