#pragma once

#include <managed_window.hpp>
#include <static_control.hpp>
#include <font.hpp>
#include <brush.hpp>

#include <wfsession.hpp>

#include <sstream>
#include <map>


class control_window : public managed_window {
private:

	wfsession _session;
	font _font;

	std::map<int, static_control> _controls;

	static const int CtrlId_Resolution = 12346;
	static const int CtrlId_MainGeometry = 12347;
	static const int CtrlId_OverlayGeometry = 12348;
	static const int CtrlId_Placeholder = 12349;

	brush _brush;

public:

	/// constructor. default constructor.
	control_window()
		: managed_window(NULL, null_deleter)
		, _font(nullptr)
		, _brush(brush::solid_brush(RGB(255, 255, 255))) {
	}

	/// constructor.
	control_window(HWND hwnd, DeleterFn&& deleter, const wfsession& session)
		: managed_window(hwnd, std::move(deleter))
		, _session(session)
		, _font(nullptr)
		, _brush(brush::solid_brush(RGB(255, 255, 255))) {

		_controls.emplace(CtrlId_Resolution, static_control::create(hwnd, CtrlId_Resolution, L"Resolution", {0, 0, 200, 14}));
		_controls.emplace(CtrlId_MainGeometry, static_control::create(hwnd, CtrlId_MainGeometry, L"MainWindow Geometry", {200, 0, 600, 14}));
		_controls.emplace(CtrlId_OverlayGeometry, static_control::create(hwnd, CtrlId_OverlayGeometry, L"Overlay Geometry: ? x ?", {200, 14, 600, 28}));
		_controls.emplace(CtrlId_Placeholder, static_control::create(hwnd, CtrlId_Placeholder, L"", {0, 14, 200, 28}));

		NONCLIENTMETRICS ncm = {};
		ncm.cbSize = sizeof(ncm);
		if (SystemParametersInfo(SPI_GETNONCLIENTMETRICS, sizeof(ncm), &ncm, 0)) {
			if (font statusFont = font::create(&ncm.lfStatusFont); statusFont.valid()) {
				_font = std::move(statusFont);
				for (auto& control : _controls) {
					control.second.send_message(WM_SETFONT, (WPARAM)(HFONT)_font, 0);
				}

				update_layout(); // update layout since font changed
			}
		}

		auto settings = _session.query_client_display();
		on_display_changed(settings.first.cx, settings.first.cy);
		on_position_changed(); // temp

		for (auto& control : _controls) {
			control.second.show();
		}
	}

	/// move constructor.
	control_window(control_window&& other) noexcept
		: managed_window(std::move(other))
		, _font(nullptr)
		, _brush(brush::solid_brush(RGB(255, 255, 255))) {

		std::swap(_session, other._session);
		std::swap(_font, other._font);
		std::swap(_controls, other._controls);
		std::swap(_brush, other._brush);
	}

	/// move assignment operator.
	control_window& operator =(control_window&& other) noexcept {
		if (this != &other) {
			managed_window::operator =(std::move(other));
			std::swap(_session, other._session);
			std::swap(_font, other._font);
			std::swap(_controls, other._controls);
			std::swap(_brush, other._brush);
		}
		return *this;
	}

	/// main window factory method
	static control_window create(HINSTANCE hInstance, const std::wstring& title, const wfsession& session) {
		using window_class_type = window_class<control_window>;
		static window_class_type wndclass = window_class_type::register_class(hInstance, L"ctxoverlay_control_window_class", CreateSolidBrush(RGB(255, 255, 255)));

		return managed_window::create(wndclass, title, null_deleter, session);
	}

	void with_control(int id, std::function<void(static_control&)> fn) {
		if (auto ctrl = _controls.find(id); ctrl != _controls.end()) {
			fn(ctrl->second);
		}
	}

	std::pair<LONG, LONG> measure_control(int id) {
		std::pair<LONG, LONG> result{};
		with_control(id, [this, &result](static_control& control) { result = measure_control(control); });
		return result;
	}

	std::pair<LONG, LONG> measure_control(static_control& control) const {
		SIZE size{};
		control.get_device_context().using_object((HFONT)_font, [&control, &size](device_context& dc) {
			(void)GetTextExtentPoint32(dc, control.get_text().c_str(), (int)control.get_text().size(), &size);
			});

		return {size.cx, size.cy};
	}

	void offset_control_vertically(int id, long amount) {
		reposition_control_by_id(id, [amount](RECT& rc) {
			auto h = rc.bottom - rc.top;
			rc.top = amount;
			rc.bottom = rc.top + h;
			});
	}

	void offset_control_horizontally(int id, long amount) {
		reposition_control_by_id(id, [amount](RECT& rc) {
			auto w = rc.right - rc.left;
			rc.left = amount;
			rc.right = rc.left + w;
			});
	}

	void reposition_control_by_id(int id, std::function<void(RECT&)> placer) {
		with_control(id, [this, &placer](static_control& control) { reposition_control(control, placer); });
	}

	void reposition_control(static_control& control, std::function<void(RECT&)> placer) {
			RECT rc{};
			control.get_window_rect(&rc);
			MapWindowPoints(HWND_DESKTOP, _hwnd, (LPPOINT)&rc, 2);
			placer(rc);
			control.move_window(rc);
	}

	void resize_control(int id, long cx, long cy) {
		reposition_control_by_id(id, [cx, cy](RECT& rc) {
			rc.right = rc.left + cx;
			rc.bottom = rc.top + cy;
			});
	}


	void update_layout() {

		auto [resw, resh] = measure_control(CtrlId_Resolution);

		resw += 20; // add 20 pixels for padding

		auto default_height = compute_default_control_height();

		// second row
		offset_control_vertically(CtrlId_OverlayGeometry, default_height);
		offset_control_vertically(CtrlId_Placeholder, default_height);

		// second column
		offset_control_horizontally(CtrlId_MainGeometry, resw);
		offset_control_horizontally(CtrlId_OverlayGeometry, resw);

		RECT rcWin{};
		get_window_rect(&rcWin);
		auto window_width = rcWin.right - rcWin.left;

		resize_control(CtrlId_MainGeometry, window_width - resw, default_height);
		resize_control(CtrlId_OverlayGeometry, window_width - resw, default_height);

		resize_control(CtrlId_Resolution, resw, default_height);
		resize_control(CtrlId_Placeholder, resw, default_height);
	}

	long compute_default_control_height() const {
		auto temp = static_control::create(_hwnd, 12350, L"Test", { 0, 0, 100, 14 });
		auto [width, height] = measure_control(temp);
		temp.destroy();
		return height;
	}

	unsigned int get_height() const {
		return compute_default_control_height() * 2;
	}

	void update_overlay_geometry(window_message& msg) {
		std::wostringstream ss;
		ss << "Overlay Geometry: (";
		ss << msg.pos.x;
		ss << ", ";
		ss << msg.pos.y;
		ss << ") ";
		ss << msg.size.dx;
		ss << " x ";
		ss << msg.size.dy;

		if (auto control = _controls.find(CtrlId_OverlayGeometry); control != _controls.end()) {
			control->second.set_text(ss.str());
		}
	}

	void update_borders(RECT rc) {
		std::wostringstream ss;
		ss << "Borders: (";
		ss << rc.left;
		ss << ", ";
		ss << rc.top;
		ss << ") (";
		ss << rc.right;
		ss << ", ";
		ss << rc.bottom;
		ss << ")";

		if (auto control = _controls.find(CtrlId_Placeholder); control != _controls.end()) {
			control->second.set_text(ss.str());
		}
	}

private:

	void on_position_changed() {

		RECT rc = {};
		get_parent().get_window_rect(&rc);

		std::wostringstream ss;
		ss << "MainWindow Geometry: (";
		ss << rc.left; // pt.x;
		ss << ", ";
		ss << rc.top; //pt.y;
		ss << ") ";
		ss << rc.right - rc.left;
		ss << " x ";
		ss << rc.bottom - rc.top;

		if (auto control = _controls.find(CtrlId_MainGeometry); control != _controls.end()) {
			control->second.set_text(ss.str());
		}
	}

	void on_display_changed(int dx, int dy) {

		auto settings = _session.query_client_display();
		std::wostringstream ss;
		ss << "Display Resolution: ";
		ss << dx;
		ss << " x ";
		ss << dy;

		if (auto control = _controls.find(CtrlId_Resolution); control != _controls.end()) {
			control->second.set_text(ss.str());
		}

		update_layout();
	}

	LRESULT on_color_static(bool& handled) {
		handled = true;
		return (LRESULT)(HBRUSH)_brush;
	}

	LRESULT on_message(UINT message, WPARAM wparam, LPARAM lparam, bool& handled) {
		switch (message) {
		case WM_USER:
		case WM_MOVE:
			on_position_changed();
			break;

		case WM_DISPLAYCHANGE:
			on_display_changed(LOWORD(lparam), HIWORD(lparam));
			break;

		case WM_CTLCOLORSTATIC:
			return on_color_static(handled);

		default:
			break;
		}
		return 0;
	}
};