#pragma once

#include "content.hpp"
#include "remote_window.hpp"

#include "overlay_session.hpp"


/// overlayed content. refines the display content concept in two ways: keeps track of an overlay window and provides an event callback for loading the content on the endpoint.
/// real content is rendered by the local endpoint device and overlayed on top of the session to appear normal.
template <typename session_type, const unsigned int message_id>
class overlayed_content : public display_content<remote_window> {
private:

	overlay_session::unique_resource _client;
	session_type _session;
	window _message_window;

public:

	/// constructor. construct with a window handle, for which the image data is to be retrieved, a parent window handle, to which events will be posted, and an overlay window, in which the content will be displayed.
	overlayed_content(typename session_type::key_type id, const window& message_window, remote_window&& remote)
		: display_content((std::size_t)id, std::move(remote))
		, _message_window(message_window) {
	}

	/// destructor.
	~overlayed_content() {

		// tell the client to stop capturing this app before we tell the server to stop broadcasting it.
		_client.reset();
	}

	/// copy constructor. copy construction is not allowed because each class instance owns the underlying resource(s) exclusively.
	overlayed_content(const overlayed_content&) = delete;

	/// copy assignment operator. copy assignment is not allowed for the same reason that copy construction is not allowed.
	overlayed_content& operator =(const overlayed_content&) = delete;

	/// move constructor. move construction is not allowed because the underlying resource(s) hold a pointer to the object through which events are delivered.
	overlayed_content(overlayed_content&&) = default;

	/// move assignment operator. move assignment is not allowed for the same reason that move construction is not allowed.
	overlayed_content& operator =(overlayed_content&&) = default;

	void render(HWND hwndMsg, UINT message_code, UINT32 flags) override {

		//_client.reset();

		typename session_type::key_type key = (typename session_type::key_type)id();
		_session = session_type::create(key, bind_events(key, hwndMsg, message_code), flags);
	}

	void on_event(session_event evt, const unsigned int& id) override {
		switch (evt) {
		case session_event::started:
			on_display_started(id);
			break;

		case session_event::stopped:
			on_display_stopped(id);
			break;

		case session_event::updated:
			break;

		default:
			break;
		}
	}

private:

	static auto subscriber(const window& message_window, UINT message_code) {
		return [message_window, message_code](std::size_t id, UINT_PTR data) {
			message_window.send_message_to_main_thread(message_code, (WPARAM)data, (LPARAM)id);
		};
	}

	void on_display_started(const unsigned int& id) {

		_client = overlay_session::make_unique(subscriber(_message_window, message_id), id, target().id());
	}

	void on_display_stopped(const unsigned int& id) {
		//_client.reset();
	}

	session_callback_type bind_events(typename session_type::key_type session_key, const window& message_window, UINT message_code) {
		return [message_window, message_code, session_key](session_event evt, const unsigned int& id) {
			// use post message here because events may be received on a worker thread and we want everything
			// to be processed on the main message loop thread to avoid synchronization issues.
			return message_window.send_message_to_main_thread(message_code + (unsigned int)evt, (WPARAM)id, (LPARAM)session_key);
		};
	}
};