#pragma once

#include <functional>



class overlay_session {
private:

	using callback_type = std::function<void(const std::size_t&, UINT_PTR)>;
	callback_type _fn;
	unsigned int _id{ 0 }; /// the overlay session identifier.
	unsigned int _overlay_id{ 0 }; /// the overlay window identifier.


	/// control the session. starts or stops an active session.
	void control_session(bool start) {
		unsigned int i = _overlay_id << 1;
		if (start) i |= 0x1;
		if (_fn) _fn(_id, i);
	}

	/// start a new client session.
	static overlay_session open(callback_type&& callback, const unsigned int& id, const unsigned int& overlay_id) {
		overlay_session session{ std::move(callback), id, overlay_id };
		session.control_session(true);
		return session;
	}

	/// close the client session.
	static void close(overlay_session& session) {
		session.control_session(false);
	}

	/// a "deleter" type for overlay sessions that can be used with std::unique_ptr.
	struct release {
		using pointer = overlay_session;
		void operator ()(pointer session) const { close(session); }
	};

public:

	/// a unique client session type.
	using unique_resource = std::unique_ptr<overlay_session, release>;

	/// make a unique client session.
	static unique_resource make_unique(callback_type&& callback, const unsigned int& id, const unsigned int& overlay_id) {
		return unique_resource(open(std::move(callback), id, overlay_id));
	}

	/// constructor. default constructor.
	overlay_session() = default;

	/// constructor. construct with overlay and session id.
	overlay_session(callback_type&& callback, const unsigned int& id, const unsigned int& overlay_id)
		: _fn(std::move(callback))
		, _id(id)
		, _overlay_id(overlay_id) {
	}

	/// constructor. construct from null pointer. provided to meet the NullablePointer requirements so that overlay_session can be used with std::unique_ptr.
	overlay_session(std::nullptr_t) {}

	/// boolean conversion operator.
	explicit operator bool() const noexcept { return (bool)_fn; }

	/// return the overlay session id.
	const unsigned int& id() const { return _id; }
};